# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import abc


class TargetNotFoundError(Exception):
    @abc.abstractproperty
    def _target_type(self):
        return None

    def __init__(self, *args, **kwargs):
        self._target = kwargs.pop("target", None)

        super(TargetNotFoundError, self).__init__(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        item_list = [Exception.__str__(self, *args, **kwargs)]

        if self._target:
            item_list.append("{} not found: {}".format(self._target_type, self._target))

        return " ".join(item_list).strip()

    def __repr__(self, *args, **kwargs):
        return self.__str__(*args, **kwargs)


class NetworkInterfaceNotFoundError(TargetNotFoundError):
    """
    Exception raised when network interface not found.
    """

    @property
    def _target_type(self):
        return "network interface"

    def __str__(self, *args, **kwargs):
        item_list = [super(NetworkInterfaceNotFoundError, self).__str__(*args, **kwargs)]

        try:
            import netifaces

            item_list.append("(available interfaces: {})".format(", ".join(netifaces.interfaces())))
        except ImportError:
            pass

        return " ".join(item_list).strip()


class ContainerNotFoundError(TargetNotFoundError):
    """
    Exception raised when container not found.
    """

    @property
    def _target_type(self):
        return "container"

    def __str__(self, *args, **kwargs):
        from ._docker import DockerClient

        dclient = DockerClient()
        container_names = dclient.extract_running_container_names()
        item_list = [super(ContainerNotFoundError, self).__str__(*args, **kwargs)]

        if container_names:
            item_list.append(
                "(available running containers: {})".format(", ".join(container_names))
            )
        else:
            item_list.append("(running container not found)")

        return " ".join(item_list).strip()


class ModuleNotFoundError(Exception):
    """
    Exception raised when mandatory kernel module not found.
    """


class TcCommandExecutionError(Exception):
    """
    Exception raised when failed to execute a ``tc`` command.
    """


class TcAlreadyExist(TcCommandExecutionError):
    """
    Exception raised when a traffic shaping rule already exist.
    """


class ParameterError(ValueError):
    """
    Exception raised when an invalid parameter specified for
    a traffic shaping rule.
    """

    def __init__(self, *args, **kwargs):
        self.__value = kwargs.pop("value", None)
        self.__expected = kwargs.pop("expected", None)

        super(ParameterError, self).__init__(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        item_list = [ValueError.__str__(self, *args, **kwargs)]
        extra_msg_list = self._get_extra_msg_list()

        if extra_msg_list:
            item_list.extend([":", ", ".join(extra_msg_list)])

        return " ".join(item_list)

    def __repr__(self, *args, **kwargs):
        return self.__str__(*args, **kwargs)

    def _get_extra_msg_list(self):
        extra_msg_list = []

        if self.__expected:
            extra_msg_list.append("expected={}".format(self.__expected))

        if self.__value:
            extra_msg_list.append("value={}".format(self.__value))

        return extra_msg_list


class UnitNotFoundError(ParameterError):
    """
    """

    def __init__(self, *args, **kwargs):
        self.__available_unit = kwargs.pop("available_unit", None)

        super(UnitNotFoundError, self).__init__(*args, **kwargs)

    def _get_extra_msg_list(self):
        extra_msg_list = []

        if self.__available_unit:
            extra_msg_list.append("available-units={}".format(self.__available_unit))

        return super(UnitNotFoundError, self)._get_extra_msg_list() + extra_msg_list



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, division, unicode_literals

import re
from collections import namedtuple

import typepy

from ._error import ParameterError, UnitNotFoundError
from ._logger import logger


ByteUnit = namedtuple("ByteUnit", "regexp factor")

_RE_NUMBER = re.compile(r"^[-\+]?[0-9\.]+")


class HumanReadableBits(object):
    __UNITS = [
        ByteUnit(regexp=re.compile("^b$", re.IGNORECASE), factor=0),
        ByteUnit(regexp=re.compile("^bps$", re.IGNORECASE), factor=0),
        ByteUnit(regexp=re.compile("^k$", re.IGNORECASE), factor=1),
        ByteUnit(regexp=re.compile("^kbps$", re.IGNORECASE), factor=1),
        ByteUnit(regexp=re.compile("^m$", re.IGNORECASE), factor=2),
        ByteUnit(regexp=re.compile("^mbps$", re.IGNORECASE), factor=2),
        ByteUnit(regexp=re.compile("^g$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^gbps$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^t$", re.IGNORECASE), factor=4),
        ByteUnit(regexp=re.compile("^tbps$", re.IGNORECASE), factor=4),
    ]

    def __init__(self, readable_size, kilo_size=1024):
        """
        String converter that human-readable byte size to a number.

        :param str readable_size: human readable size (bytes). e.g. 256 M
        :param int kilo_size: size of ``kilo``. 1024 or 1000
        """

        self.__readable_size = readable_size
        self.__kilo_size = kilo_size

        VALID_KILLO_SIZE = (1000, 1024)
        if self.__kilo_size not in VALID_KILLO_SIZE:
            raise ParameterError(
                "invalid kilo size", expected=VALID_KILLO_SIZE, value=self.__kilo_size
            )

        self.__bits = None

    def to_bits(self):
        """
        :raises ParameterError:
        """

        if self.__bits is not None:
            return self.__bits

        logger.debug("human readable size to bit: {}".format(self.__readable_size))

        typepy.String(self.__readable_size).validate(error_message="readable_size must be a string")

        self.__readable_size = self.__readable_size.strip()

        try:
            size = _RE_NUMBER.search(self.__readable_size).group()
        except AttributeError:
            raise ParameterError("readable_size require number", value=self.__readable_size)
        size = float(size)
        if size < 0:
            raise ParameterError("size must be greater or equals to zero", value=size)

        unit = _RE_NUMBER.sub("", self.__readable_size).strip().lower()

        self.__bits = size * self.__get_coefficient(unit)

        return self.__bits

    def to_kilobits(self):
        """
        :param str readable_size: human readable size (bytes). e.g. 256 M
        :raises ValueError:
        """

        return self.to_bits() / self.__kilo_size

    def __get_coefficient(self, unit_str):
        for unit in self.__UNITS:
            if unit.regexp.search(unit_str):
                return self.__kilo_size ** unit.factor

        raise UnitNotFoundError(
            "unit not found", value=unit_str, available_unit="b/bps/k/kbps/m/mbps/g/gbps/t/tbps"
        )


class HumanReadableTime(object):
    __VALID_MINUTE_UNITS = ["m", "min", "mins", "minute", "minutes"]
    __VALID_SEC_UNITS = ["s", "sec", "secs", "second", "seconds"]
    __VALID_MSEC_UNITS = ["ms", "msec", "msecs", "millisecond", "milliseconds"]
    __VALID_USEC_UNITS = ["us", "usec", "usecs", "microsecond", "microseconds"]
    __VALID_UNIT_LIST = (
        __VALID_MINUTE_UNITS + __VALID_SEC_UNITS + __VALID_MSEC_UNITS + __VALID_USEC_UNITS
    )

    def __init__(self, readable_time):
        self.__readable_time = readable_time

    def __eq__(self, other):
        return self.get_msec() == other.get_msec()

    def __ne__(self, other):
        return self.get_msec() != other.get_msec()

    def __lt__(self, other):
        return self.get_msec() < other.get_msec()

    def __le__(self, other):
        return self.get_msec() <= other.get_msec()

    def __gt__(self, other):
        return self.get_msec() > other.get_msec()

    def __ge__(self, other):
        return self.get_msec() >= other.get_msec()

    def __repr__(self):
        return self.get_value()

    @classmethod
    def get_valid_units(cls):
        return cls.__VALID_UNIT_LIST

    def get_value(self):
        self.__preprocess()

        value_fmt = "{:f}{:s}"
        if typepy.Integer(self.__number, strict_level=typepy.StrictLevel.MAX).is_type():
            value_fmt = "{:d}{:s}"

        return value_fmt.format(self.__number, self.__unit)

    def get_msec(self):
        self.__preprocess()

        coef = 1
        if self.__unit == "sec":
            coef = 1000
        elif self.__unit == "us":
            coef = 0.001

        return self.__number * coef

    def validate(self, min_value=None, max_value=None):
        if min_value is not None:
            if not isinstance(min_value, HumanReadableTime):
                min_value = HumanReadableTime(min_value)

            if self < min_value:
                raise ParameterError(
                    "time value is too low",
                    expected="greater than or equal to {}".format(min_value),
                    value=self,
                )

        if max_value is not None:
            if not isinstance(max_value, HumanReadableTime):
                max_value = HumanReadableTime(max_value)

            if self > max_value:
                raise ParameterError(
                    "time value is too high",
                    expected="less than or equal to {}".format(max_value),
                    value=self,
                )

        if self.__unit not in self.__VALID_UNIT_LIST:
            raise UnitNotFoundError(
                "unknown unit",
                value=self.__readable_time,
                available_unit="/".join(self.__VALID_UNIT_LIST),
            )

    def __normalize(self):
        number = typepy.Integer(self.__number).try_convert()
        if number is not None:
            self.__number = number

        if self.__unit in self.__VALID_SEC_UNITS:
            self.__unit = "sec"
        elif self.__unit in self.__VALID_MSEC_UNITS:
            self.__unit = "ms"
        elif self.__unit in self.__VALID_USEC_UNITS:
            self.__unit = "us"
        elif self.__unit in self.__VALID_MINUTE_UNITS:
            self.__number *= 60
            self.__unit = "sec"

    def __get_number(self):
        match = _RE_NUMBER.search(self.__readable_time)
        if not match:
            raise ParameterError(
                "human-readable time must be include a number", value=self.__readable_time
            )

        return float(match.group())

    def __get_unit(self):
        if typepy.RealNumber(self.__readable_time).is_type():
            # if the input value is real numbers consider unit as milliseconds.
            return "ms"

        return _RE_NUMBER.sub("", self.__readable_time).strip().lower()

    def __preprocess(self):
        self.__number = self.__get_number()
        self.__unit = self.__get_unit()
        self.validate()
        self.__normalize()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import pytest

import tcconfig._common
from tcconfig._const import TcSubCommand
from tcconfig._tc_command_helper import get_tc_base_command


class Test_get_tc_base_command(object):
    @pytest.mark.parametrize(
        ["subcommand", "expected"],
        [
            [TcSubCommand.CLASS, "/sbin/tc class"],
            [TcSubCommand.FILTER, "/sbin/tc filter"],
            [TcSubCommand.QDISC, "/sbin/tc qdisc"],
        ],
    )
    def test_normal(self, monkeypatch, subcommand, expected):
        monkeypatch.setattr(tcconfig._common, "find_bin_path", lambda _: "/sbin/tc")

        assert get_tc_base_command(subcommand) == expected

    @pytest.mark.parametrize(
        ["subcommand", "expected"], [["qdisc", ValueError], [None, ValueError]]
    )
    def test_exception(self, subcommand, expected):
        with pytest.raises(expected):
            get_tc_base_command(subcommand)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, print_function

import sys

from subprocrunner import SubprocessRunner
from typepy import RealNumber

from tcconfig._const import Tc
from tcconfig._converter import HumanReadableBits, HumanReadableTime


DEADLINE_TIME = 3  # [sec]
ASSERT_MARGIN = 0.5


def print_test_result(expected, actual, error=None):
    print("[expected]\n{}\n".format(expected))
    print("[actual]\n{}\n".format(actual))

    if error:
        print(error, file=sys.stderr)


def is_invalid_param(rate, delay, packet_loss, packet_duplicate, corrupt, reordering):
    param_value_list = [packet_loss, packet_duplicate, corrupt, reordering]

    print(param_value_list)

    is_invalid = all(
        [
            not RealNumber(param_value).is_type() or param_value <= 0
            for param_value in param_value_list
        ]
        + [HumanReadableTime(delay) <= HumanReadableTime("0ms")]
    )

    try:
        HumanReadableBits(rate, kilo_size=1000).to_bits()
    except (TypeError, ValueError):
        pass
    else:
        is_invalid = False

    return is_invalid


def execute_tcdel(device):
    return SubprocessRunner([Tc.Command.TCDEL, device, "--all"], dry_run=False).run()



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, print_function, unicode_literals

import sys

import msgfy
import simplejson as json
import subprocrunner as spr
from docker.errors import DockerException

from .__version__ import __version__
from ._argparse_wrapper import ArgparseWrapper
from ._common import check_command_installation, initialize_cli
from ._const import Tc, TcCommandOutput
from ._docker import DockerClient
from ._error import TargetNotFoundError
from ._logger import logger
from ._network import verify_network_interface
from ._tc_script import write_tc_script
from .parser.shaping_rule import TcShapingRuleParser


def parse_option():
    parser = ArgparseWrapper(__version__)

    group = parser.parser.add_argument_group("Traffic Control")
    if set(["-d", "--device"]).intersection(set(sys.argv)):
        # deprecated: remain for backward compatibility
        group.add_argument(
            "-d", "--device", action="append", required=True, help="network device name (e.g. eth0)"
        )
    else:
        group.add_argument("device", nargs="+", help="network device name (e.g. eth0)")
    group.add_argument(
        "--ipv6",
        dest="ip_version",
        action="store_const",
        const=6,
        default=4,
        help="Display IPv6 shaping rules. Defaults to show IPv4 shaping rules.",
    )

    parser.add_docker_group(is_add_srcdst=False)

    parser.parser.add_argument(
        "--color",
        action="store_true",
        default=False,
        help="colorize the output. require Pygments package.",
    )

    return parser.parser.parse_args()


def print_tc(text, is_colorize):
    try:
        from pygments import highlight
        from pygments.lexers import JsonLexer
        from pygments.formatters import TerminalTrueColorFormatter

        pygments_installed = True
    except ImportError:
        pygments_installed = False

    if is_colorize and pygments_installed:
        print(
            highlight(
                code=text, lexer=JsonLexer(), formatter=TerminalTrueColorFormatter(style="monokai")
            )
        )
    else:
        print(text)


def extract_tc_params(options):
    dclient = None
    if options.use_docker:
        try:
            dclient = DockerClient(options.tc_command_output)
        except DockerException as e:
            logger.error(msgfy.to_error_message(e))
            sys.exit(1)

    tc_params = {}

    for device in options.device:
        try:
            if options.use_docker and dclient.exist_container(container=device):
                container = device

                dclient.verify_container(container)
                dclient.create_veth_table(container)
                container_info = dclient.extract_container_info(container)

                for veth in dclient.fetch_veth_list(container_info.name):
                    tc_params.update(
                        TcShapingRuleParser(
                            veth, options.ip_version, options.tc_command_output, logger
                        ).get_tc_parameter()
                    )
                    key = "{veth} (container_id={id}, image={image})".format(
                        veth=veth, id=container_info.id[:12], image=container_info.image
                    )
                    tc_params[key] = tc_params.pop(veth)
            else:
                verify_network_interface(device, options.tc_command_output)

                tc_params.update(
                    TcShapingRuleParser(
                        device, options.ip_version, options.tc_command_output, logger
                    ).get_tc_parameter()
                )
        except TargetNotFoundError as e:
            logger.warn(e)
            continue

    return tc_params


def main():
    options = parse_option()

    initialize_cli(options)
    check_command_installation("tc")

    if options.tc_command_output != TcCommandOutput.NOT_SET:
        spr.SubprocessRunner.default_is_dry_run = True

    tc_params = extract_tc_params(options)
    command_history = "\n".join(spr.SubprocessRunner.get_history())

    if options.tc_command_output == TcCommandOutput.STDOUT:
        print(command_history)
        return 0

    if options.tc_command_output == TcCommandOutput.SCRIPT:
        write_tc_script(
            Tc.Command.TCSHOW, command_history, filename_suffix="-".join(options.device)
        )
        return 0

    logger.debug("command history\n{}".format(command_history))

    print_tc(json.dumps(tc_params, ensure_ascii=False, indent=4), options.color)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import unicode_literals

import random

import pytest

from tcconfig._iptables import (
    VALID_CHAIN_LIST,
    IptablesMangleController,
    IptablesMangleMarkEntry,
    get_iptables_base_command,
)


_DEF_SRC = "192.168.0.0/24"
_DEF_DST = "192.168.100.0/24"


prerouting_mangle_mark_list = [
    IptablesMangleMarkEntry(
        ip_version=4,
        line_number=1,
        mark_id=1,
        source=_DEF_SRC,
        destination=_DEF_DST,
        chain="PREROUTING",
        protocol="all",
    )
]
input_mangle_mark_list = [
    IptablesMangleMarkEntry(
        ip_version=4,
        line_number=1,
        mark_id=1234,
        source="anywhere",
        destination=_DEF_DST,
        chain="INPUT",
        protocol="all",
    )
]
output_mangle_mark_list = [
    IptablesMangleMarkEntry(
        ip_version=4,
        line_number=1,
        mark_id=12,
        source=_DEF_SRC,
        destination=_DEF_DST,
        chain="OUTPUT",
        protocol="tcp",
    ),
    IptablesMangleMarkEntry(
        ip_version=4,
        line_number=2,
        mark_id=123,
        source=_DEF_SRC,
        destination="anywhere",
        chain="OUTPUT",
        protocol="all",
    ),
    IptablesMangleMarkEntry(
        ip_version=4,
        line_number=3,
        mark_id=12345,
        source="anywhere",
        destination="anywhere",
        chain="OUTPUT",
        protocol="all",
    ),
]
mangle_mark_list = prerouting_mangle_mark_list + input_mangle_mark_list + output_mangle_mark_list
reverse_mangle_mark_list = (
    list(reversed(prerouting_mangle_mark_list))
    + list(reversed(input_mangle_mark_list))
    + list(reversed(output_mangle_mark_list))
)


@pytest.fixture
def iptables_ctrl_ipv4():
    return IptablesMangleController(True, ip_version=4)


class Test_IptablesMangleMark_repr(object):
    def test_smoke(self):
        for mangle_mark in mangle_mark_list:
            assert len(str(mangle_mark)) > 0


class Test_IptablesMangleMark_to_append_command(object):
    _CMD_PREFIX = get_iptables_base_command() + " -A {:s} -t mangle -j MARK"

    @pytest.mark.parametrize(
        ["mark_id", "source", "destination", "chain", "protocol", "line_number", "expected"],
        [
            [
                2,
                _DEF_SRC,
                _DEF_DST,
                "PREROUTING",
                "all",
                None,
                "{} --set-mark 2 -p all -s {} -d {}".format(
                    _CMD_PREFIX.format("PREROUTING"), _DEF_SRC, _DEF_DST
                ),
            ],
            [
                2,
                _DEF_SRC,
                _DEF_DST,
                "OUTPUT",
                "all",
                1,
                "{} --set-mark 2 -p all -s {} -d {}".format(
                    _CMD_PREFIX.format("OUTPUT"), _DEF_SRC, _DEF_DST
                ),
            ],
            [
                2,
                _DEF_SRC,
                _DEF_DST,
                "OUTPUT",
                "tcp",
                1,
                "{} --set-mark 2 -p tcp -s {} -d {}".format(
                    _CMD_PREFIX.format("OUTPUT"), _DEF_SRC, _DEF_DST
                ),
            ],
            [
                100,
                _DEF_SRC,
                "anywhere",
                "INPUT",
                "all",
                100,
                "{} --set-mark 100 -p all -s {}".format(_CMD_PREFIX.format("INPUT"), _DEF_SRC),
            ],
            [
                1,
                "anywhere",
                _DEF_DST,
                "OUTPUT",
                "all",
                100,
                "{} --set-mark 1 -p all -d {}".format(_CMD_PREFIX.format("OUTPUT"), _DEF_DST),
            ],
            [
                1,
                "anywhere",
                "anywhere",
                "OUTPUT",
                "all",
                100,
                "{} --set-mark 1 -p all".format(_CMD_PREFIX.format("OUTPUT")),
            ],
        ],
    )
    def test_normal(self, mark_id, source, destination, chain, protocol, line_number, expected):
        mark = IptablesMangleMarkEntry(
            ip_version=4,
            mark_id=mark_id,
            source=source,
            destination=destination,
            chain=chain,
            protocol=protocol,
            line_number=line_number,
        )
        assert mark.to_append_command() == expected


class Test_IptablesMangleMark_to_delete_command(object):
    @pytest.mark.parametrize(
        ["mark_id", "source", "destination", "chain", "protocol", "line_number", "expected"],
        [
            [
                2,
                _DEF_SRC,
                _DEF_DST,
                "PREROUTING",
                "all",
                1,
                "{:s} -t mangle -D PREROUTING 1".format(get_iptables_base_command()),
            ],
            [
                20,
                None,
                None,
                "OUTPUT",
                "all",
                2,
                "{:s} -t mangle -D OUTPUT 2".format(get_iptables_base_command()),
            ],
        ],
    )
    def test_normal(self, mark_id, source, destination, chain, protocol, line_number, expected):
        mark = IptablesMangleMarkEntry(
            ip_version=4,
            mark_id=mark_id,
            source=_DEF_SRC,
            destination=_DEF_DST,
            chain=chain,
            protocol=protocol,
            line_number=line_number,
        )
        assert mark.to_delete_command() == expected

    @pytest.mark.parametrize(
        ["mark_id", "source", "destination", "chain", "protocol", "line_number", "expected"],
        [[2, _DEF_SRC, _DEF_DST, "OUTPUT", "all", None, TypeError]],
    )
    def test_exception(self, mark_id, source, destination, chain, protocol, line_number, expected):
        mark = IptablesMangleMarkEntry(
            ip_version=4,
            mark_id=mark_id,
            source=source,
            destination=destination,
            chain=chain,
            protocol=protocol,
            line_number=line_number,
        )
        with pytest.raises(expected):
            mark.to_delete_command()


class Test_IptablesMangleController_get_unique_mark_id(object):
    @pytest.mark.xfail(run=False)
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()

        for i in range(5):
            mark_id = iptables_ctrl_ipv4.get_unique_mark_id()

            assert mark_id == (i + 101)

            mangle_mark = IptablesMangleMarkEntry(
                ip_version=4,
                mark_id=mark_id,
                source=_DEF_SRC,
                destination=_DEF_DST,
                chain=random.choice(VALID_CHAIN_LIST),
            )

            assert iptables_ctrl_ipv4.add(mangle_mark) == 0


class Test_IptablesMangleController_add(object):
    @pytest.mark.xfail(run=False)
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()
        initial_len = len(iptables_ctrl_ipv4.get_iptables())

        for mangle_mark in mangle_mark_list:
            assert iptables_ctrl_ipv4.add(mangle_mark) == 0

        assert len(iptables_ctrl_ipv4.get_iptables()) > initial_len


class Test_IptablesMangleController_clear(object):
    @pytest.mark.xfail(run=False)
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()

        initial_len = len(iptables_ctrl_ipv4.get_iptables())

        for mangle_mark in mangle_mark_list:
            assert iptables_ctrl_ipv4.add(mangle_mark) == 0

        assert len(iptables_ctrl_ipv4.get_iptables()) > initial_len

        iptables_ctrl_ipv4.clear()

        assert len(iptables_ctrl_ipv4.get_iptables()) == initial_len


class Test_IptablesMangleController_parse(object):
    @pytest.mark.xfail(run=False)
    def test_normal(self, iptables_ctrl_ipv4):
        iptables_ctrl_ipv4.clear()

        for mangle_mark in mangle_mark_list:
            assert iptables_ctrl_ipv4.add(mangle_mark) == 0

        for lhs_mangle, rhs_mangle in zip(iptables_ctrl_ipv4.parse(), reverse_mangle_mark_list):
            print("lhs: {}".format(lhs_mangle))
            print("rhs: {}".format(rhs_mangle))

            assert lhs_mangle == rhs_mangle



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, print_function, unicode_literals

import errno
import os
import re
import sys
from collections import namedtuple

import msgfy
import six
from docker import APIClient
from docker.errors import APIError, NotFound
from path import Path
from simplesqlite import OperationalError, connect_memdb
from simplesqlite.model import Integer, Model, Text
from simplesqlite.query import And, Where
from subprocrunner import SubprocessRunner

from ._common import is_execute_tc_command
from ._const import TcCommandOutput
from ._error import ContainerNotFoundError
from ._logger import logger


if six.PY2:
    PermissionError = OSError


ContainerInfo = namedtuple("ContainerInfo", "id name pid ipaddr image")


class IfIndex(Model):
    host = Text(not_null=True)
    ifindex = Integer(primary_key=True)
    ifname = Text(not_null=True)
    peer_ifindex = Integer(not_null=True, unique=True)


class DockerClient(object):
    @property
    def __netns_root_path(self):
        return Path("/var/run/netns")

    def __init__(self, tc_command_output=TcCommandOutput.NOT_SET):
        self.__client = APIClient(version="auto")
        self.__host_name = os.uname()[1]
        self.__tc_command_output = tc_command_output

        self.__con = connect_memdb()
        IfIndex.attach(self.__con)

    def exist_container(self, container):
        try:
            self.__verify_container(container)
            return True
        except ContainerNotFoundError:
            return False

    def verify_container(self, container, exit_on_exception=False):
        if not is_execute_tc_command(self.__tc_command_output):
            return

        try:
            self.__verify_container(container)
        except ContainerNotFoundError as e:
            if exit_on_exception:
                logger.error(msgfy.to_error_message(e))
                sys.exit(errno.EPERM)

            raise

    def extract_running_container_names(self):
        running_container_name_list = []

        for container in self.__get_containers():
            if container.get("State") != "running":
                continue

            running_container_name_list.append(container["Names"][0].lstrip("/"))

        return running_container_name_list

    def extract_container_info(self, container):
        try:
            container_map = self.__client.inspect_container(container=container)
        except APIError as e:
            logger.error(e)
            sys.exit(1)

        container_name = container_map["Name"].lstrip("/")
        container_state = container_map["State"]

        if not container_state["Running"]:
            logger.error("{} not running".format(container_name))
            return ContainerInfo(name=container_name)

        return ContainerInfo(
            id=container_map["Id"],
            name=container_name,
            pid=int(container_state["Pid"]),
            ipaddr=container_map["NetworkSettings"]["IPAddress"],
            image=container_map["Config"]["Image"],
        )

    def create_veth_table(self, container):
        try:
            self.__netns_root_path.makedirs_p()
        except PermissionError as e:
            logger.error(e)
            sys.exit(errno.EPERM)

        container_info = self.extract_container_info(container)
        logger.debug(
            "found container: name={}, pid={}".format(container_info.name, container_info.pid)
        )

        try:
            netns_path = self.__get_netns_path(container_info.name)

            if not os.path.lexists(netns_path):
                logger.debug("make symlink to {}".format(netns_path))

                try:
                    Path("/proc/{:d}/ns/net".format(container_info.pid)).symlink(netns_path)
                except PermissionError as e:
                    logger.error(e)
                    sys.exit(errno.EPERM)

            return_code = self.__create_ifindex_table(container_info.name)
            if return_code != 0:
                sys.exit(return_code)

            return return_code
        finally:
            netns_path.remove_p()

    def select_veth(self, container_name):
        for container_record in IfIndex.select(where=Where("host", container_name)):
            for veth_record in IfIndex.select(
                where=And(
                    [
                        Where("host", self.__host_name),
                        Where("ifindex", container_record.peer_ifindex),
                    ]
                )
            ):
                yield veth_record

    def fetch_veth_list(self, container_name):
        return [veth_record.ifname for veth_record in self.select_veth(container_name)]

    def __verify_container(self, container):
        if len(self.__get_containers()) == 0:
            raise ContainerNotFoundError()

        try:
            self.__client.inspect_container(container=container)
        except NotFound:
            raise ContainerNotFoundError(target=container)
        except APIError as e:
            logger.error(e)
            sys.exit(1)

    def __get_containers(self):
        try:
            return self.__client.containers()
        except APIError as e:
            logger.error(e)
            sys.exit(1)

    def __get_netns_path(self, container_name):
        return self.__netns_root_path / container_name

    def __create_ifindex_table(self, container_name):
        netns_path = self.__get_netns_path(container_name)

        try:
            netns_path.stat()
        except PermissionError as e:
            logger.error(e)
            return errno.EPERM

        IfIndex.create()

        proc = SubprocessRunner(
            "ip netns exec {ns} ip link show type veth".format(ns=container_name), dry_run=False
        )
        if proc.run() != 0:
            logger.error(proc.stderr)
            return proc.returncode

        veth_groups_regexp = re.compile("([0-9]+): ([a-z0-9]+)@([a-z0-9]+): ")
        peer_ifindex_prefix_regexp = re.compile("^if")

        try:
            for i, line in enumerate(proc.stdout.splitlines()):
                match = veth_groups_regexp.search(line)
                if not match:
                    continue

                logger.debug("parse veth @{} [{:02d}] {}".format(container_name, i, line))
                ifindex, ifname, peer_ifindex = match.groups()
                IfIndex.insert(
                    IfIndex(
                        host=container_name,
                        ifindex=int(ifindex),
                        ifname=ifname,
                        peer_ifindex=int(peer_ifindex_prefix_regexp.sub("", peer_ifindex)),
                    )
                )

            proc = SubprocessRunner("ip link show type veth", dry_run=False)
            if proc.run() != 0:
                logger.error(proc.stderr)
                return proc.returncode

            for line in proc.stdout.splitlines():
                logger.debug("parse veth @docker-host [{:02d}] {}".format(i, line))
                match = veth_groups_regexp.search(line)
                if not match:
                    continue

                ifindex, ifname, peer_ifindex = match.groups()
                try:
                    IfIndex.insert(
                        IfIndex(
                            host=self.__host_name,
                            ifindex=int(ifindex),
                            ifname=ifname,
                            peer_ifindex=int(peer_ifindex_prefix_regexp.sub("", peer_ifindex)),
                        )
                    )
                except OperationalError as e:
                    logger.error(msgfy.to_error_message(e))
        finally:
            IfIndex.commit()

            try:
                netns_path.remove_p()
            except PermissionError as e:
                logger.error(msgfy.to_error_message(e))
                return errno.EPERM

        return 0



# encoding: utf-8

from __future__ import absolute_import, unicode_literals

import sys

import msgfy
from docker.errors import DockerException

from ._const import TcCommandOutput
from ._docker import DockerClient
from ._logger import logger
from ._tc_script import write_tc_script


class Main(object):
    def __init__(self, options):
        self._options = options

        self._dclient = None
        if self._options.use_docker:
            try:
                self._dclient = DockerClient(options.tc_command_output)
            except DockerException as e:
                logger.error(msgfy.to_error_message(e))
                sys.exit(1)

    def _extract_dst_network(self):
        if self._options.dst_container:
            return self._dclient.extract_container_info(self._options.dst_container).ipaddr

        return self._options.dst_network

    def _extract_src_network(self):
        if self._options.src_container:
            return self._dclient.extract_container_info(self._options.src_container).ipaddr

        return self._options.src_network

    def _fetch_tc_targets(self):
        if not self._options.use_docker:
            return [self._options.device]

        container = self._options.device

        self._dclient.verify_container(container, exit_on_exception=True)
        self._dclient.create_veth_table(container)

        return self._dclient.fetch_veth_list(self._dclient.extract_container_info(container).name)

    def _get_return_code(self, return_code_list):
        error_return_code = None

        for return_code in return_code_list:
            if return_code == 0:
                return return_code

            error_return_code = return_code

        return error_return_code

    def _dump_history(self, tc, tc_command):
        command_history = "\n".join(tc.get_command_history())
        command_output = self._options.tc_command_output

        if command_output == TcCommandOutput.STDOUT:
            print(command_history)
            return

        try:
            filename_suffix = tc.netem_param.make_param_name()
        except AttributeError:
            filename_suffix = None

        if command_output == TcCommandOutput.SCRIPT:
            write_tc_script(tc_command, command_history, filename_suffix=filename_suffix)
            return

        logger.debug("command history\n{}".format(command_history))



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import division

import pytest

from tcconfig._converter import HumanReadableBits, HumanReadableTime
from tcconfig._error import ParameterError, UnitNotFoundError


class Test_to_bit(object):
    @pytest.mark.parametrize(
        ["value", "kilo_size", "expected"],
        [
            ["2b", 1024, 2],
            ["2 b", 1024, 2],
            ["2 B", 1024, 2],
            ["2k", 1024, 2 * 1024 ** 1],
            ["2 k", 1024, 2 * 1024 ** 1],
            ["2 K", 1024, 2 * 1024 ** 1],
            ["2m", 1024, 2 * 1024 ** 2],
            ["2 m", 1024, 2 * 1024 ** 2],
            ["2 M", 1024, 2 * 1024 ** 2],
            ["2g", 1024, 2 * 1024 ** 3],
            ["2 g", 1024, 2 * 1024 ** 3],
            ["2 G", 1024, 2 * 1024 ** 3],
            ["2t", 1024, 2 * 1024 ** 4],
            ["2 t", 1024, 2 * 1024 ** 4],
            ["2 T", 1024, 2 * 1024 ** 4],
            ["2.5 M", 1024, int(2.5 * 1024 ** 2)],
            ["2.5 m", 1024, int(2.5 * 1024 ** 2)],
            ["2b", 1000, 2],
            ["2 b", 1000, 2],
            ["2 B", 1000, 2],
            ["2k", 1000, 2 * 1000 ** 1],
            ["2 k", 1000, 2 * 1000 ** 1],
            ["2 K", 1000, 2 * 1000 ** 1],
            ["2m", 1000, 2 * 1000 ** 2],
            ["2 m", 1000, 2 * 1000 ** 2],
            ["2 M", 1000, 2 * 1000 ** 2],
            ["2g", 1000, 2 * 1000 ** 3],
            ["2 g", 1000, 2 * 1000 ** 3],
            ["2 G", 1000, 2 * 1000 ** 3],
            ["2t", 1000, 2 * 1000 ** 4],
            ["2 t", 1000, 2 * 1000 ** 4],
            ["2 T", 1000, 2 * 1000 ** 4],
            ["2.5 M", 1000, int(2.5 * 1000 ** 2)],
            ["2.5 m", 1000, int(2.5 * 1000 ** 2)],
        ],
    )
    def test_normal(self, value, kilo_size, expected):
        assert HumanReadableBits(value, kilo_size).to_bits() == expected

    @pytest.mark.parametrize(
        ["value", "kilo_size", "exception"],
        [
            ["10", 1000, UnitNotFoundError],
            [None, 1000, TypeError],
            [True, 1000, TypeError],
            [float("nan"), 1000, TypeError],
            ["", 1000, ParameterError],
            ["a", 1000, ParameterError],
            ["1k0 ", 1000, ParameterError],
            ["10kb", 1000, ParameterError],
            ["-2m", 1000, ParameterError],
            ["2m", None, ParameterError],
            ["2m", 1001, ParameterError],
        ],
    )
    def test_exception(self, value, kilo_size, exception):
        with pytest.raises(exception):
            HumanReadableBits(value, kilo_size).to_bits()


class Test_to_kilo_bit(object):
    @pytest.mark.parametrize(
        ["value", "kilo_size", "expected"],
        [
            ["2b", 1024, 2 / 1024],
            ["2k", 1024, 2 * 1024 ** 0],
            ["2m", 1024, 2 * 1024 ** 1],
            ["2g", 1024, 2 * 1024 ** 2],
            ["2t", 1024, 2 * 1024 ** 3],
            ["2b", 1000, 2 / 1000],
            ["2k", 1000, 2 * 1000 ** 0],
            ["2m", 1000, 2 * 1000 ** 1],
            ["2g", 1000, 2 * 1000 ** 2],
            ["2t", 1000, 2 * 1000 ** 3],
        ],
    )
    def test_normal(self, value, kilo_size, expected):
        assert HumanReadableBits(value, kilo_size).to_kilobits() == expected


class Test_HumanReadableTime_get_value(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            ["10", "10ms"],
            ["1s", "1sec"],
            ["1S", "1sec"],
            ["1sec", "1sec"],
            ["0.1secs", "0.100000sec"],
            ["1.5second", "1.500000sec"],
            ["1Seconds", "1sec"],
            ["11ms", "11ms"],
            ["11.5MS", "11.500000ms"],
            ["11msec", "11ms"],
            ["11Msecs", "11ms"],
            ["11millisecond", "11ms"],
            ["11milliseconds", "11ms"],
            ["123us", "123us"],
            ["123US", "123us"],
            ["123usec", "123us"],
            ["123Usecs", "123us"],
            ["123microsecond", "123us"],
            ["123microseconds", "123us"],
            ["0.1m", "6.000000sec"],
            ["0.2M", "12.000000sec"],
            ["1.1min", "66.000000sec"],
            ["0.1Mins", "6.000000sec"],
            ["0.1minute", "6.000000sec"],
            ["0.1Minutes", "6.000000sec"],
        ],
    )
    def test_normal(self, value, expected):
        assert HumanReadableTime(value).get_value() == expected
        assert str(HumanReadableTime(value)) == expected

    @pytest.mark.parametrize(
        ["value", "exception"],
        [
            ["", ParameterError],
            [None, TypeError],
            [True, TypeError],
            [float("nan"), TypeError],
            ["a", ParameterError],
            ["1k0 ", ParameterError],
            ["10kb", ParameterError],
        ],
    )
    def test_exception(self, value, exception):
        with pytest.raises(exception):
            HumanReadableTime(value).get_value()


class Test_HumanReadableTime_get_msec(object):
    @pytest.mark.parametrize(
        ["value", "expected"], [["10", 10], ["1s", 1000], ["1ms", 1], ["1000us", 1], ["1m", 60000]]
    )
    def test_normal(self, value, expected):
        assert HumanReadableTime(value).get_msec() == expected


class Test_HumanReadableTime_validate(object):
    @pytest.mark.parametrize(
        ["value", "min_value", "max_value"],
        [
            ["1", "0", "60m"],
            ["1s", "0s", "60m"],
            ["1s", HumanReadableTime("0s"), HumanReadableTime("60m")],
            ["1s", "1s", "60m"],
            ["10ms", "0s", "60m"],
            ["100us", "0s", "60m"],
        ],
    )
    def test_normal(self, value, min_value, max_value):
        HumanReadableTime(value).validate(min_value=min_value, max_value=max_value)

    @pytest.mark.parametrize(
        ["value", "min_value", "max_value", "expected"],
        [
            ["1s", "0s", "1ms", ParameterError],
            ["-1s", "0s", "60m", ParameterError],
            ["10ms", "0s", "1ms", ParameterError],
            ["-10ms", "0s", "60m", ParameterError],
            ["100us", "0s", "60us", ParameterError],
            ["-100us", "0s", "60m", ParameterError],
        ],
    )
    def test_exception(self, value, min_value, max_value, expected):
        with pytest.raises(expected):
            HumanReadableTime(value).validate(min_value=min_value, max_value=max_value)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import typepy
from subprocrunner import SubprocessRunner

from ._interface import AbstractShaper
from .._common import logging_context, run_command_helper
from .._const import ShapingAlgorithm, TcSubCommand, TrafficDirection
from .._error import ParameterError
from .._network import get_anywhere_network, get_no_limit_kbits


class TbfShaper(AbstractShaper):

    __NETEM_QDISC_MAJOR_ID_OFFSET = 10

    __MIN_BUFFER_BYTE = 1600

    __OUT_DEVICE_QDISC_MINOR_ID = 1
    __IN_DEVICE_QDISC_MINOR_ID = 3

    @property
    def algorithm_name(self):
        return ShapingAlgorithm.TBF

    def _get_qdisc_minor_id(self):
        if self._tc_obj.direction == TrafficDirection.OUTGOING:
            return self.__OUT_DEVICE_QDISC_MINOR_ID

        if self._tc_obj.direction == TrafficDirection.INCOMING:
            return self.__IN_DEVICE_QDISC_MINOR_ID

        raise ParameterError(
            "unknown direction", expected=TrafficDirection.LIST, value=self._tc_obj.direction
        )

    def _get_netem_qdisc_major_id(self, base_id):
        if self._tc_obj.direction == TrafficDirection.OUTGOING:
            direction_offset = 0
        elif self._tc_obj.direction == TrafficDirection.INCOMING:
            direction_offset = 1

        return base_id + self.__NETEM_QDISC_MAJOR_ID_OFFSET + direction_offset

    def _make_qdisc(self):
        base_command = self._tc_obj.get_tc_command(TcSubCommand.QDISC)
        handle = "{:s}:".format(self._tc_obj.qdisc_major_id_str)

        return run_command_helper(
            " ".join([base_command, self._dev, "root", "handle {:s}".format(handle), "prio"]),
            ignore_error_msg_regexp=self._tc_obj.REGEXP_FILE_EXISTS,
            notice_msg=self._tc_obj.EXISTS_MSG_TEMPLATE.format(
                "failed to '{command:s}': prio qdisc already exists "
                "({dev:s}, algo={algorithm:s}, handle={handle:s}).".format(
                    command=base_command,
                    dev=self._dev,
                    algorithm=self.algorithm_name,
                    handle=handle,
                )
            ),
        )

    def _add_rate(self):
        try:
            self._tc_obj.netem_param.validate_bandwidth_rate()
        except ParameterError:
            return 0

        base_command = self._tc_obj.get_tc_command(TcSubCommand.QDISC)
        parent = "{:x}:{:d}".format(
            self._get_netem_qdisc_major_id(self._tc_obj.qdisc_major_id), self._get_qdisc_minor_id()
        )
        handle = "{:d}:".format(20)
        no_limit_kbits = get_no_limit_kbits(self._tc_device)

        kbits = self._tc_obj.netem_param.bandwidth_rate
        if kbits is None:
            kbits = no_limit_kbits

        command = " ".join(
            [
                base_command,
                self._dev,
                "parent {:s}".format(parent),
                "handle {:s}".format(handle),
                self.algorithm_name,
                "rate {}kbit".format(kbits),
                "buffer {:d}".format(max(int(kbits), self.__MIN_BUFFER_BYTE)),  # [byte]
                "limit 10000",
            ]
        )

        run_command_helper(
            command,
            ignore_error_msg_regexp=self._tc_obj.REGEXP_FILE_EXISTS,
            notice_msg=self._tc_obj.EXISTS_MSG_TEMPLATE.format(
                "failed to '{command:s}': qdisc already exists ({dev:s}, algo={algorithm:s}, "
                "parent={parent:s}, handle={handle:s}).".format(
                    command=base_command,
                    dev=self._dev,
                    algorithm=self.algorithm_name,
                    parent=parent,
                    handle=handle,
                )
            ),
        )

        self.__set_pre_network_filter()

    def set_shaping(self):
        with logging_context("_make_qdisc"):
            self._make_qdisc()

        with logging_context("_set_netem"):
            self._set_netem()

        with logging_context("_add_rate"):
            self._add_rate()

        with logging_context("_add_filter"):
            self._add_filter()

        return 0

    def __set_pre_network_filter(self):
        if self._is_use_iptables():
            return 0

        if all(
            [
                typepy.is_null_string(self._tc_obj.dst_network),
                not typepy.Integer(self._tc_obj.dst_port).is_type(),
            ]
        ):
            flowid = "{:s}:{:d}".format(self._tc_obj.qdisc_major_id_str, self._get_qdisc_minor_id())
        else:
            flowid = "{:s}:2".format(self._tc_obj.qdisc_major_id_str)

        return SubprocessRunner(
            " ".join(
                [
                    self._tc_obj.get_tc_command(TcSubCommand.FILTER),
                    self._dev,
                    "protocol {:s}".format(self._tc_obj.protocol),
                    "parent {:s}:".format(self._tc_obj.qdisc_major_id_str),
                    "prio 2 u32 match {:s} {:s} {:s}".format(
                        self._tc_obj.protocol,
                        self._get_network_direction_str(),
                        get_anywhere_network(self._tc_obj.ip_version),
                    ),
                    "flowid {:s}".format(flowid),
                ]
            )
        ).run()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import division, print_function, unicode_literals

import pytest
import typepy
from allpairspy import AllPairs
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc

from .common import execute_tcdel


SKIP_TEST = False


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


def is_valid_combination(row):
    if all([typepy.is_null_string(param) for param in row]):
        return False

    return True


def is_invalid_param(rate, delay, loss, corrupt):
    params = [rate, delay, loss, corrupt]

    return all([typepy.is_null_string(param) for param in params])


class NormalTestValue(object):
    RATE_LIST = ["", "--rate 100K", "--rate 0.5M"]
    DELAY_LIST = ["", "--delay 100ms"]
    DELAY_DISTRO_LIST = ["", "--delay-distro 20ms"]
    PACKET_LOSS_RATE_LIST = ["", "--loss 0.1", "--loss 10%"]
    CORRUPTION_RATE_LIST = ["", "--corrupt 0.1", "--loss 10%"]
    DIRECTION_LIST = ["", "--direction outgoing", "--direction incoming"]
    NETWORK_LIST = ["", "--network 192.168.0.10", "--network 192.168.0.0/24"]
    PORT_LIST = ["", "--port 80"]
    OVERWRITE_LIST = ["", "--add", "--overwrite"]
    IPTABLES_LIST = ["", "--iptables"]


class Test_tcconfig(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:
      python setup.py test --addopts "--runxfail --device=<test device>"

    These tests are expected to execute on following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.skipif("SKIP_TEST is True")
    @pytest.mark.parametrize(
        [
            "rate",
            "delay",
            "delay_distro",
            "loss",
            "corrupt",
            "direction",
            "network",
            "port",
            "overwrite",
            # "is_enable_iptables",
        ],
        [
            opt_list
            for opt_list in AllPairs(
                [
                    NormalTestValue.RATE_LIST,
                    NormalTestValue.DELAY_LIST,
                    NormalTestValue.DELAY_DISTRO_LIST,
                    NormalTestValue.PACKET_LOSS_RATE_LIST,
                    NormalTestValue.CORRUPTION_RATE_LIST,
                    NormalTestValue.DIRECTION_LIST,
                    NormalTestValue.NETWORK_LIST,
                    NormalTestValue.PORT_LIST,
                    NormalTestValue.OVERWRITE_LIST,
                    # NormalTestValue.IPTABLES_LIST,
                ],
                n=3,
                filter_func=is_valid_combination,
            )
        ],
    )
    def test_smoke(
        self,
        device_value,
        rate,
        delay,
        delay_distro,
        loss,
        corrupt,
        direction,
        network,
        port,
        overwrite,  # is_enable_iptables
    ):

        if device_value is None:
            pytest.skip("device is empty")

        if is_invalid_param(rate, delay, loss, corrupt):
            pytest.skip("skip null parameters")

        for device_option in [device_value, "--device {}".format(device_value)]:
            execute_tcdel(device_value)

            command = " ".join(
                [
                    Tc.Command.TCSET,
                    device_option,
                    rate,
                    delay,
                    delay_distro,
                    loss,
                    corrupt,
                    direction,
                    network,
                    port,
                    overwrite,
                    # is_enable_iptables,
                ]
            )
            print("command: {}".format(command))
            tcset_proc = SubprocessRunner(command)
            assert tcset_proc.run() == 0, tcset_proc.stderr

            execute_tcdel(device_value)



# encoding: utf-8

from datetime import datetime


__author__ = "Tsuyoshi Hombashi"
__copyright__ = "Copyright 2016-{}, {}".format(datetime.now().year, __author__)
__license__ = "MIT License"
__version__ = "0.21.9"
__maintainer__ = __author__
__email__ = "tsuyoshi.hombashi@gmail.com"



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import copy

import logbook
import simplesqlite
import subprocrunner
import typepy
from typepy import Integer

from ._class import TcClassParser
from ._filter import TcFilterParser
from ._qdisc import TcQdiscParser
from .._common import is_execute_tc_command
from .._const import Tc, TcSubCommand, TrafficDirection
from .._error import NetworkInterfaceNotFoundError
from .._iptables import IptablesMangleController
from .._network import is_anywhere_network
from .._tc_command_helper import get_tc_base_command, run_tc_show


class TcShapingRuleParser(object):
    @property
    def con(self):
        return self.__con

    @property
    def device(self):
        return self.__device

    @property
    def ifb_device(self):
        return self.__ifb_device

    def __init__(self, device, ip_version, tc_command_output, logger):
        self.__device = device
        self.__ip_version = ip_version
        self.__tc_command_output = tc_command_output
        self.__logger = logger

        self.clear()
        self.__ifb_device = self.__get_ifb_from_device()

        self.__iptables_ctrl = IptablesMangleController(True, ip_version)

    def clear(self):
        self.__con = simplesqlite.connect_memdb()
        self.__filter_parser = TcFilterParser(self.__con, self.__ip_version)
        self.__parsed_mappings = {}

    def get_tc_parameter(self):
        return {
            self.device: {
                TrafficDirection.OUTGOING: self.__get_shaping_rule(self.device),
                TrafficDirection.INCOMING: self.__get_shaping_rule(self.ifb_device),
            }
        }

    def get_outgoing_tc_filter(self):
        return self.__parse_tc_filter(self.device)

    def get_incoming_tc_filter(self):
        if not self.ifb_device:
            return []

        return self.__parse_tc_filter(self.ifb_device)

    def parse(self):
        self.__parse_device(self.device)
        self.__parse_device(self.ifb_device)

    def __parse_device(self, device):
        if not device:
            return

        if self.__parsed_mappings.get(device):
            return

        self.__parse_tc_class(device)
        self.__parse_tc_filter(device)
        self.__parse_tc_qdisc(device)

        self.__parsed_mappings[device] = True

    def __get_ifb_from_device(self):
        if not is_execute_tc_command(self.__tc_command_output):
            return None

        filter_runner = subprocrunner.SubprocessRunner(
            "{:s} show dev {:s} root".format(get_tc_base_command(TcSubCommand.FILTER), self.device),
            error_log_level=logbook.NOTSET,
            dry_run=False,
        )
        if filter_runner.run() != 0 and filter_runner.stderr.find("Cannot find device") != -1:
            raise NetworkInterfaceNotFoundError(target=self.device)

        return self.__filter_parser.parse_incoming_device(filter_runner.stdout)

    def __get_filter_key(self, filter_param):
        src_network_format = Tc.Param.SRC_NETWORK + "={:s}"
        dst_network_format = Tc.Param.DST_NETWORK + "={:s}"
        protocol_format = Tc.Param.PROTOCOL + "={:s}"
        key_item_list = []

        if Tc.Param.HANDLE in filter_param:
            handle = filter_param.get(Tc.Param.HANDLE)
            Integer(handle).validate()
            handle = int(handle)

            for mangle in self.__iptables_ctrl.parse():
                if mangle.mark_id != handle:
                    continue

                key_item_list.append(dst_network_format.format(mangle.destination))
                if typepy.is_not_null_string(mangle.source):
                    key_item_list.append("{:s}={:s}".format(Tc.Param.SRC_NETWORK, mangle.source))
                key_item_list.append(protocol_format.format(mangle.protocol))

                break
            else:
                raise ValueError("mangle mark not found: {}".format(mangle))
        else:
            src_network = filter_param.get(Tc.Param.SRC_NETWORK)
            if typepy.is_not_null_string(src_network) and not is_anywhere_network(
                src_network, self.__ip_version
            ):
                key_item_list.append(src_network_format.format(src_network))

            dst_network = filter_param.get(Tc.Param.DST_NETWORK)
            if typepy.is_not_null_string(dst_network) and not is_anywhere_network(
                dst_network, self.__ip_version
            ):
                key_item_list.append(dst_network_format.format(dst_network))

            src_port = filter_param.get(Tc.Param.SRC_PORT)
            if Integer(src_port).is_type():
                port_format = Tc.Param.SRC_PORT + "={:d}"
                key_item_list.append(port_format.format(src_port))

            dst_port = filter_param.get(Tc.Param.DST_PORT)
            if Integer(dst_port).is_type():
                port_format = Tc.Param.DST_PORT + "={:d}"
                key_item_list.append(port_format.format(dst_port))

            protocol = filter_param.get(Tc.Param.PROTOCOL)
            if typepy.is_not_null_string(protocol):
                key_item_list.append(protocol_format.format(protocol))

        return ", ".join(key_item_list)

    def __get_shaping_rule(self, device):
        from simplesqlite.query import Where

        if typepy.is_null_string(device):
            return {}

        self.__parse_device(device)
        where_query = Where(Tc.Param.DEVICE, device)

        try:
            class_param_list = self.__con.select_as_dict(
                table_name=TcSubCommand.CLASS.value, where=where_query
            )
        except simplesqlite.TableNotFoundError:
            class_param_list = []

        try:
            filter_param_list = self.__con.select_as_dict(
                table_name=TcSubCommand.FILTER.value, where=where_query
            )
        except simplesqlite.TableNotFoundError:
            filter_param_list = []

        try:
            qdisc_param_list = self.__con.select_as_dict(
                table_name=TcSubCommand.QDISC.value, where=where_query
            )
        except simplesqlite.TableNotFoundError:
            qdisc_param_list = []

        shaping_rule_mapping = {}

        for filter_param in filter_param_list:
            self.__logger.debug("{:s} param: {}".format(TcSubCommand.FILTER, filter_param))
            shaping_rule = {}

            filter_key = self.__get_filter_key(filter_param)
            if typepy.is_null_string(filter_key):
                self.__logger.debug("empty filter key: {}".format(filter_param))
                continue

            for qdisc_param in qdisc_param_list:
                self.__logger.debug("{:s} param: {}".format(TcSubCommand.QDISC, qdisc_param))

                if qdisc_param.get(Tc.Param.PARENT) not in (
                    filter_param.get(Tc.Param.FLOW_ID),
                    filter_param.get(Tc.Param.CLASS_ID),
                ):
                    continue

                shaping_rule[Tc.Param.FILTER_ID] = filter_param.get(Tc.Param.FILTER_ID)
                # shaping_rule[Tc.Param.PRIORITY] = filter_param.get(
                #    Tc.Param.PRIORITY)
                shaping_rule.update(
                    self.__strip_param(
                        qdisc_param, [Tc.Param.DEVICE, Tc.Param.PARENT, Tc.Param.HANDLE]
                    )
                )

            for class_param in class_param_list:
                self.__logger.debug("{:s} param: {}".format(TcSubCommand.CLASS, class_param))

                if class_param.get(Tc.Param.CLASS_ID) not in (
                    filter_param.get(Tc.Param.FLOW_ID),
                    filter_param.get(Tc.Param.CLASS_ID),
                ):
                    continue

                shaping_rule[Tc.Param.FILTER_ID] = filter_param.get(Tc.Param.FILTER_ID)
                # shaping_rule[Tc.Param.PRIORITY] = filter_param.get(
                #    Tc.Param.PRIORITY)
                shaping_rule.update(
                    self.__strip_param(class_param, [Tc.Param.DEVICE, Tc.Param.CLASS_ID])
                )

            if not shaping_rule:
                self.__logger.debug("shaping rule not found for '{}'".format(filter_param))
                continue

            self.__logger.debug("shaping rule found: {} {}".format(filter_key, shaping_rule))

            shaping_rule_mapping[filter_key] = shaping_rule

        return shaping_rule_mapping

    def __parse_tc_qdisc(self, device):
        return TcQdiscParser(self.__con).parse(
            device, run_tc_show(TcSubCommand.QDISC, device, self.__tc_command_output)
        )

    def __parse_tc_filter(self, device):
        return self.__filter_parser.parse(
            device, run_tc_show(TcSubCommand.FILTER, device, self.__tc_command_output)
        )

    def __parse_tc_class(self, device):
        return TcClassParser(self.__con).parse(
            device, run_tc_show(TcSubCommand.CLASS, device, self.__tc_command_output)
        )

    @staticmethod
    def __strip_param(params, strip_param_list):
        work_params = copy.deepcopy(params)

        for strip_param in strip_param_list:
            try:
                del work_params[strip_param]
            except KeyError:
                pass

        return work_params



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import re

import simplejson as json
import typepy

from ._interface import AbstractParser
from .._const import ShapingAlgorithm, Tc, TcSubCommand
from .._logger import logger


class TcClassParser(AbstractParser):
    class Pattern(object):
        CLASS_ID = "[0-9a-z:]+"
        RATE = "[0-9]+[KMGT]?"

    class Key(object):
        DEVICE = Tc.Param.DEVICE
        CLASS_ID = Tc.Param.CLASS_ID
        RATE = "rate"

        LIST = (DEVICE, CLASS_ID, RATE)

    @property
    def _tc_subcommand(self):
        return TcSubCommand.CLASS.value

    def __init__(self, con):
        super(TcClassParser, self).__init__()

        self.__con = con

    def parse(self, device, text):
        entry_list = []

        for line in text.splitlines():
            self._clear()

            if typepy.is_null_string(line):
                continue

            line = self._to_unicode(line.lstrip())

            self.__parsed_param[self.Key.DEVICE] = device
            self.__parse_classid(line)
            self.__parse_rate(line)

            logger.debug("parse a class entry: {}".format(self.__parsed_param))
            entry_list.append(self.__parsed_param)

        if entry_list:
            self.__con.create_table_from_data_matrix(self._tc_subcommand, self.Key.LIST, entry_list)

        logger.debug(
            "tc {:s} parse result: {}".format(self._tc_subcommand, json.dumps(entry_list, indent=4))
        )

        return entry_list

    def _clear(self):
        self.__parsed_param = {}

    def __parse_classid(self, line):
        self.__parsed_param[self.Key.CLASS_ID] = None
        tag = "class {:s} ".format(ShapingAlgorithm.HTB)

        match = re.search("{:s}{:s}".format(tag, self.Pattern.CLASS_ID), line)
        if match is None:
            return

        self.__parsed_param[self.Key.CLASS_ID] = re.search(
            self.Pattern.CLASS_ID, match.group().lstrip(tag)
        ).group()

    def __parse_rate(self, line):
        self.__parsed_param[self.Key.RATE] = None

        match = re.search("rate {:s}".format(self.Pattern.RATE), line)
        if match is None:
            return

        self.__parsed_param[self.Key.RATE] = (
            re.search(self.Pattern.RATE, match.group()).group() + "bps"
        )



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <>
"""

from __future__ import absolute_import, unicode_literals

import errno
import os
import re
import sys

import subprocrunner as spr

from ._common import find_bin_path
from ._logger import logger


def get_required_capabilities(command):
    required_capabilities_map = {
        "tc": ["cap_net_admin"],
        "ip": ["cap_net_raw", "cap_net_admin"],
        "iptables": ["cap_net_raw", "cap_net_admin"],
    }

    return required_capabilities_map[command]


def get_permission_error_message(command):
    PERMISSION_ERROR_MSG_FORMAT = "\n".join(
        [
            "Permission denied: you must be root or set Linux capabilities to execute the command.",
            "  How to setup Linux capabilities for the {command:s} command:",
            "    $ sudo setcap {capabilities:s}+ep {bin_path:s}",
        ]
    )

    return PERMISSION_ERROR_MSG_FORMAT.format(
        command=command,
        capabilities=",".join(get_required_capabilities(command)),
        bin_path=find_bin_path(command),
    )


def _has_capabilies(bin_path, capabilities):
    getcap_bin_path = find_bin_path("getcap")

    if not getcap_bin_path:
        logger.error("command not found: getcap")
        return False

    bin_path = os.path.realpath(bin_path)
    proc = spr.SubprocessRunner("{:s} {:s}".format(getcap_bin_path, bin_path))
    if proc.run() != 0:
        logger.error(proc.stderr)
        sys.exit(proc.returncode)

    getcap_output = proc.stdout
    has_capabilies = True
    for capability in capabilities:
        if re.search(capability, getcap_output):
            logger.debug("{:s} has {:s} capability".format(bin_path, capability))
        else:
            logger.debug("{:s} has no {:s} capability".format(bin_path, capability))
            has_capabilies = False

    capability = "+ep"
    if re.search(re.escape(capability), getcap_output):
        logger.debug("{:s} has {:s} capability".format(bin_path, capability))
    else:
        logger.debug("{:s} has no {:s} capability".format(bin_path, capability))
        has_capabilies = False

    return has_capabilies


def has_execution_authority(command):
    from ._common import check_command_installation

    check_command_installation(command)

    if os.getuid() == 0:
        return True

    return _has_capabilies(find_bin_path(command), get_required_capabilities(command))


def check_execution_authority(command):
    if has_execution_authority(command):
        return

    logger.error(get_permission_error_message(command))
    sys.exit(errno.EPERM)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, division, unicode_literals

import errno
import re

import typepy

from ._interface import AbstractShaper
from .._common import is_execute_tc_command, logging_context, run_command_helper
from .._const import ShapingAlgorithm, TcSubCommand
from .._error import TcAlreadyExist
from .._logger import logger
from .._network import get_no_limit_kbits
from .._tc_command_helper import run_tc_show


class HtbShaper(AbstractShaper):
    __DEFAULT_CLASS_MINOR_ID = 1

    class MinQdiscMinorId(object):
        OUTGOING = 40
        INCOMING = 20

    @property
    def algorithm_name(self):
        return ShapingAlgorithm.HTB

    def __init__(self, tc_obj):
        super(HtbShaper, self).__init__(tc_obj)

        self.__qdisc_minor_id = None
        self.__qdisc_minor_id_count = 0
        self.__netem_major_id = None
        self.__classid_wo_shaping = None

    def _get_qdisc_minor_id(self):
        if self.__qdisc_minor_id is None:
            self.__qdisc_minor_id = self.__get_unique_qdisc_minor_id()
            logger.debug("__get_unique_qdisc_minor_id: {:d}".format(self.__qdisc_minor_id))

        return self.__qdisc_minor_id

    def _get_netem_qdisc_major_id(self, base_id):
        if self.__netem_major_id is None:
            self.__netem_major_id = self.__get_unique_netem_major_id()

        return self.__netem_major_id

    def _make_qdisc(self):
        if self._tc_obj.is_change_shaping_rule:
            return 0

        base_command = self._tc_obj.get_tc_command(TcSubCommand.QDISC)
        handle = "{:s}:".format(self._tc_obj.qdisc_major_id_str)

        if self._tc_obj.is_add_shaping_rule:
            message = None
        else:
            message = self._tc_obj.EXISTS_MSG_TEMPLATE.format(
                "failed to '{command:s}': qdisc already exists "
                "({dev:s}, handle={handle:s}, algo={algorithm:s}).".format(
                    command=base_command,
                    dev=self._dev,
                    handle=handle,
                    algorithm=self.algorithm_name,
                )
            )

        run_command_helper(
            " ".join(
                [
                    base_command,
                    self._dev,
                    "root",
                    "handle {:s}".format(handle),
                    self.algorithm_name,
                    "default {:d}".format(self.__DEFAULT_CLASS_MINOR_ID),
                ]
            ),
            ignore_error_msg_regexp=self._tc_obj.REGEXP_FILE_EXISTS,
            notice_msg=message,
            exception_class=TcAlreadyExist,
        )

        return self.__add_default_class()

    def _add_rate(self):
        base_command = self._tc_obj.get_tc_command(TcSubCommand.CLASS)
        parent = "{:s}:".format(
            self._get_tc_parent("{:s}:".format(self._tc_obj.qdisc_major_id_str)).split(":")[0]
        )
        classid = self._get_tc_parent(
            "{:s}:{:d}".format(self._tc_obj.qdisc_major_id_str, self._get_qdisc_minor_id())
        )
        no_limit_kbits = get_no_limit_kbits(self._tc_device)

        kbits = self._tc_obj.netem_param.bandwidth_rate
        if kbits is None:
            kbits = no_limit_kbits

        command_item_list = [
            base_command,
            self._dev,
            "parent {:s}".format(parent),
            "classid {:s}".format(classid),
            self.algorithm_name,
            "rate {}Kbit".format(kbits),
            "ceil {}Kbit".format(kbits),
        ]

        if kbits != no_limit_kbits:
            command_item_list.extend(
                ["burst {}KB".format(kbits / (10 * 8)), "cburst {}KB".format(kbits / (10 * 8))]
            )

        run_command_helper(
            " ".join(command_item_list),
            ignore_error_msg_regexp=self._tc_obj.REGEXP_FILE_EXISTS,
            notice_msg=self._tc_obj.EXISTS_MSG_TEMPLATE.format(
                "failed to '{command:s}': class already exists "
                "({dev:s}, parent={parent:s}, classid={classid:s}, "
                "algo={algorithm:s}).".format(
                    command=base_command,
                    dev=self._dev,
                    parent=parent,
                    classid=classid,
                    algorithm=self.algorithm_name,
                )
            ),
            exception_class=TcAlreadyExist,
        )

    def _add_exclude_filter(self):
        import subprocrunner

        if all(
            [
                typepy.is_null_string(param)
                for param in (
                    self._tc_obj.exclude_dst_network,
                    self._tc_obj.exclude_src_network,
                    self._tc_obj.exclude_dst_port,
                    self._tc_obj.exclude_src_port,
                )
            ]
        ):
            logger.debug("no exclude filter found")
            return

        command_item_list = [
            self._tc_obj.get_tc_command(TcSubCommand.FILTER),
            self._dev,
            "protocol {:s}".format(self._tc_obj.protocol),
            "parent {:s}:".format(self._tc_obj.qdisc_major_id_str),
            "prio 1",
            "u32",
        ]

        if typepy.is_not_null_string(self._tc_obj.exclude_dst_network):
            command_item_list.append(
                "match {:s} {:s} {:s}".format(
                    self._tc_obj.protocol_match, "dst", self._tc_obj.exclude_dst_network
                )
            )

        if typepy.is_not_null_string(self._tc_obj.exclude_src_network):
            command_item_list.append(
                "match {:s} {:s} {:s}".format(
                    self._tc_obj.protocol_match, "src", self._tc_obj.exclude_src_network
                )
            )

        if typepy.is_not_null_string(self._tc_obj.exclude_dst_port):
            command_item_list.append(
                "match {:s} {:s} {:s} 0xffff".format(
                    self._tc_obj.protocol_match, "dport", self._tc_obj.exclude_dst_port
                )
            )

        if typepy.is_not_null_string(self._tc_obj.exclude_src_port):
            command_item_list.append(
                "match {:s} {:s} {:s} 0xffff".format(
                    self._tc_obj.protocol_match, "sport", self._tc_obj.exclude_src_port
                )
            )

        command_item_list.append("flowid {:s}".format(self.__classid_wo_shaping))

        return subprocrunner.SubprocessRunner(" ".join(command_item_list)).run()

    def set_shaping(self):
        is_add_shaping_rule = self._tc_obj.is_add_shaping_rule

        with logging_context("_make_qdisc"):
            try:
                self._make_qdisc()
            except TcAlreadyExist:
                if not is_add_shaping_rule:
                    return errno.EINVAL

        with logging_context("_add_rate"):
            try:
                self._add_rate()
            except TcAlreadyExist:
                if not is_add_shaping_rule:
                    return errno.EINVAL

        with logging_context("_set_netem"):
            self._set_netem()

        with logging_context("_add_exclude_filter"):
            self._add_exclude_filter()

        with logging_context("_add_filter"):
            self._add_filter()

        return 0

    def __get_unique_qdisc_minor_id(self):
        if not is_execute_tc_command(self._tc_obj.tc_command_output):
            return (
                int(self._tc_obj.netem_param.calc_hash(self._tc_obj.make_srcdst_text())[-2:], 16)
                + 1
            )

        if self._tc_obj.is_change_shaping_rule:
            self.__qdisc_minor_id_count += 1

            return self.__DEFAULT_CLASS_MINOR_ID + self.__qdisc_minor_id_count

        exist_class_item_list = re.findall(
            "class {algorithm:s} {qdisc_major_id:s}:[0-9]+".format(
                algorithm=ShapingAlgorithm.HTB, qdisc_major_id=self._tc_obj.qdisc_major_id_str
            ),
            run_tc_show(TcSubCommand.CLASS, self._tc_device, self._tc_obj.tc_command_output),
            re.MULTILINE,
        )

        exist_class_minor_id_list = []
        for class_item in exist_class_item_list:
            try:
                exist_class_minor_id_list.append(typepy.Integer(class_item.split(":")[1]).convert())
            except typepy.TypeConversionError:
                continue

        logger.debug("existing class list with {:s}: {}".format(self._dev, exist_class_item_list))
        logger.debug(
            "existing minor classid list with {:s}: {}".format(self._dev, exist_class_minor_id_list)
        )

        next_minor_id = self.__DEFAULT_CLASS_MINOR_ID
        while True:
            if next_minor_id not in exist_class_minor_id_list:
                break

            next_minor_id += 1

        return next_minor_id

    def __get_unique_netem_major_id(self):
        exist_netem_item_list = re.findall(
            "qdisc [a-z]+ [a-z0-9]+",
            run_tc_show(TcSubCommand.QDISC, self._tc_device, self._tc_obj.tc_command_output),
            re.MULTILINE,
        )

        exist_netem_major_id_list = []
        for netem_item in exist_netem_item_list:
            exist_netem_major_id_list.append(int(netem_item.split()[-1], 16))

        logger.debug("existing netem list with {:s}: {}".format(self._dev, exist_netem_item_list))
        logger.debug(
            "existing netem major id list with {:s}: {}".format(
                self._dev, exist_netem_major_id_list
            )
        )

        next_netem_major_id = self._tc_obj.netem_param.calc_device_qdisc_major_id()
        while True:
            if next_netem_major_id not in exist_netem_major_id_list:
                break

            next_netem_major_id += 1

        return next_netem_major_id

    def __add_default_class(self):
        base_command = self._tc_obj.get_tc_command(TcSubCommand.CLASS)
        parent = "{:s}:".format(self._tc_obj.qdisc_major_id_str)
        classid = "{:s}:{:d}".format(self._tc_obj.qdisc_major_id_str, self.__DEFAULT_CLASS_MINOR_ID)
        self.__classid_wo_shaping = classid

        if self._tc_obj.is_add_shaping_rule:
            message = None
        else:
            message = self._tc_obj.EXISTS_MSG_TEMPLATE.format(
                "failed to default '{command:s}': class already exists "
                "({dev}, parent={parent:s}, classid={classid:s}, "
                "algo={algorithm:s})".format(
                    command=base_command,
                    dev=self._dev,
                    parent=parent,
                    classid=classid,
                    algorithm=self.algorithm_name,
                )
            )

        return run_command_helper(
            " ".join(
                [
                    base_command,
                    self._dev,
                    "parent {:s}".format(parent),
                    "classid {:s}".format(classid),
                    self.algorithm_name,
                    "rate {}kbit".format(get_no_limit_kbits(self._tc_device)),
                ]
            ),
            ignore_error_msg_regexp=self._tc_obj.REGEXP_FILE_EXISTS,
            notice_msg=message,
        )



# encoding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import pytest

from tcconfig._error import ParameterError
from tcconfig._netem_param import NetemParameter, convert_rate_to_f


class Test_convert_rate_to_f(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [[0, 0], [100.0, 100.0], ["1", 1.0], ["100", 100.0], ["0.1%", 0.1], [" 10 % ", 10.0]],
    )
    def test_normal(self, value, expected):
        assert convert_rate_to_f(value) == expected

    @pytest.mark.parametrize(
        ["value", "expected"], [["hoge", ValueError], ["1&", ValueError], ["%1", ValueError]]
    )
    def test_exception(self, value, expected):
        with pytest.raises(expected):
            convert_rate_to_f(value)


class Test_NetemParameter_validate_bandwidth_rate(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [["1bps", ParameterError], ["7bps", ParameterError], ["8bps", None], ["1Gbps", None]],
    )
    def test_normal(self, value, expected):
        param = NetemParameter(device="eth0", bandwidth_rate=value)

        if expected is None:
            param.validate_bandwidth_rate()
        else:
            with pytest.raises(expected):
                param.validate_bandwidth_rate()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import hashlib

import six
import typepy
from typepy import RealNumber

from ._common import validate_within_min_max
from ._const import KILO_SIZE, Tc
from ._converter import HumanReadableBits, HumanReadableTime
from ._error import ParameterError, UnitNotFoundError
from ._network import get_no_limit_kbits


MIN_PACKET_LOSS_RATE = 0  # [%]
MAX_PACKET_LOSS_RATE = 100  # [%]

MIN_PACKET_DUPLICATE_RATE = 0  # [%]
MAX_PACKET_DUPLICATE_RATE = 100  # [%]

MIN_CORRUPTION_RATE = 0  # [%]
MAX_CORRUPTION_RATE = 100  # [%]

MIN_REORDERING_RATE = 0  # [%]
MAX_REORDERING_RATE = 100  # [%]


def convert_rate_to_f(rate):
    if typepy.is_not_null_string(rate):
        return float(rate.rstrip("% "))

    return rate


class NetemParameter(object):
    @property
    def bandwidth_rate(self):
        # convert bandwidth string [K/M/G bit per second] to a number
        try:
            return HumanReadableBits(self.__bandwidth_rate, kilo_size=KILO_SIZE).to_kilobits()
        except (ParameterError, UnitNotFoundError, TypeError):
            return None

    def __init__(
        self,
        device,
        bandwidth_rate=None,
        latency_time=None,
        latency_distro_time=None,
        packet_loss_rate=None,
        packet_duplicate_rate=None,
        corruption_rate=None,
        reordering_rate=None,
    ):
        self.__device = device

        self.__bandwidth_rate = bandwidth_rate
        self.__latency_time = HumanReadableTime(latency_time)
        self.__latency_distro_time = HumanReadableTime(latency_distro_time)
        self.__packet_loss_rate = convert_rate_to_f(packet_loss_rate)  # [%]
        self.__packet_duplicate_rate = convert_rate_to_f(packet_duplicate_rate)  # [%]
        self.__corruption_rate = convert_rate_to_f(corruption_rate)  # [%]
        self.__reordering_rate = convert_rate_to_f(reordering_rate)  # [%]

    def validate_netem_parameter(self):
        self.validate_bandwidth_rate()
        self.__validate_network_delay()
        self.__validate_packet_loss_rate()
        self.__validate_packet_duplicate_rate()
        self.__validate_corruption_rate()
        self.__validate_reordering_rate()
        self.__validate_reordering_and_delay()

        netem_param_value_list = [
            self.bandwidth_rate,
            self.__packet_loss_rate,
            self.__packet_duplicate_rate,
            self.__corruption_rate,
            self.__reordering_rate,
        ]

        if all(
            [
                not RealNumber(netem_param_value).is_type() or netem_param_value <= 0
                for netem_param_value in netem_param_value_list
            ]
            + [self.__latency_time <= HumanReadableTime(Tc.ValueRange.LatencyTime.MIN)]
        ):
            raise ParameterError(
                "there are no valid net emulation parameters found. "
                "at least one or more following parameters are required: "
                "--rate, --delay, --loss, --duplicate, --corrupt, --reordering"
            )

    def validate_bandwidth_rate(self):
        if typepy.is_null_string(self.__bandwidth_rate):
            return

        # convert bandwidth string [K/M/G bit per second] to a number
        hr_bits = HumanReadableBits(self.__bandwidth_rate, kilo_size=KILO_SIZE)

        if hr_bits.to_bits() < 8:
            raise ParameterError(
                "bandwidth rate must be greater or equals to 8bps", value=hr_bits.to_bits()
            )

        no_limit_kbits = get_no_limit_kbits(self.__device)
        if hr_bits.to_kilobits() > no_limit_kbits:
            raise ParameterError(
                "exceed bandwidth rate limit",
                value="{} kbps".format(hr_bits.to_kilobits()),
                expected="less than {} kbps".format(no_limit_kbits),
            )

    def make_param_name(self):
        item_list = [self.__device]

        if self.bandwidth_rate:
            item_list.append("rate{}".format(self.bandwidth_rate))

        if self.__latency_time.get_msec():
            item_list.append("delay{}".format(self.__latency_time))

        if self.__latency_distro_time.get_msec():
            item_list.append("distro{}".format(self.__latency_distro_time))

        if self.__packet_loss_rate:
            item_list.append("loss{}".format(self.__packet_loss_rate))

        if self.__packet_duplicate_rate:
            item_list.append("duplicate{}".format(self.__packet_duplicate_rate))

        if self.__corruption_rate:
            item_list.append("corrupt{}".format(self.__corruption_rate))

        if self.__reordering_rate:
            item_list.append("reordering{}".format(self.__reordering_rate))

        return "_".join(item_list)

    def make_netem_command_parts(self):
        item_list = ["netem"]

        if self.__packet_loss_rate > 0:
            item_list.append("loss {:f}%".format(self.__packet_loss_rate))

        if self.__packet_duplicate_rate > 0:
            item_list.append("duplicate {:f}%".format(self.__packet_duplicate_rate))

        if self.__latency_time > HumanReadableTime(Tc.ValueRange.LatencyTime.MIN):
            item_list.append("delay {}".format(self.__latency_time))

            if self.__latency_distro_time > HumanReadableTime(Tc.ValueRange.LatencyTime.MIN):
                item_list.append("{} distribution normal".format(self.__latency_distro_time))

        if self.__corruption_rate > 0:
            item_list.append("corrupt {:f}%".format(self.__corruption_rate))

        if self.__reordering_rate > 0:
            item_list.append("reorder {:f}%".format(self.__reordering_rate))

        return " ".join(item_list)

    def calc_hash(self, extra=""):
        return hashlib.md5(six.b(self.make_param_name() + extra)).hexdigest()

    def calc_device_qdisc_major_id(self):
        base_device_hash = self.calc_hash()[:3]
        device_hash_prefix = "2"

        return int(device_hash_prefix + base_device_hash, 16)

    def __validate_network_delay(self):
        try:
            self.__latency_time.validate(
                min_value=Tc.ValueRange.LatencyTime.MIN, max_value=Tc.ValueRange.LatencyTime.MAX
            )
        except ParameterError as e:
            raise ParameterError("delay {}".format(e))

        try:
            self.__latency_distro_time.validate(
                min_value=Tc.ValueRange.LatencyTime.MIN, max_value=Tc.ValueRange.LatencyTime.MAX
            )
        except ParameterError as e:
            raise ParameterError("delay-distro {}".format(e))

    def __validate_packet_loss_rate(self):
        validate_within_min_max(
            "loss (packet loss rate)",
            self.__packet_loss_rate,
            MIN_PACKET_LOSS_RATE,
            MAX_PACKET_LOSS_RATE,
            unit="%",
        )

    def __validate_packet_duplicate_rate(self):
        validate_within_min_max(
            "duplicate (packet duplicate rate)",
            self.__packet_duplicate_rate,
            MIN_PACKET_DUPLICATE_RATE,
            MAX_PACKET_DUPLICATE_RATE,
            unit="%",
        )

    def __validate_corruption_rate(self):
        validate_within_min_max(
            "corruption (packet corruption rate)",
            self.__corruption_rate,
            MIN_CORRUPTION_RATE,
            MAX_CORRUPTION_RATE,
            unit="%",
        )

    def __validate_reordering_rate(self):
        validate_within_min_max(
            "reordering (packet reordering rate)",
            self.__reordering_rate,
            MIN_REORDERING_RATE,
            MAX_REORDERING_RATE,
            unit="%",
        )

    def __validate_reordering_and_delay(self):
        if self.__reordering_rate and self.__latency_time.get_msec() <= 0:
            raise ParameterError("reordering needs latency to be specified: set latency > 0")



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import pytest
import simplejson as json
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc

from .common import execute_tcdel, print_test_result


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcset_change(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--device=<test device>"

    These tests expected to execute in the following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    def test_smoke_multiple(self, device_value):
        if device_value is None:
            pytest.skip("device is null")

        for device_option in [device_value, "--device {}".format(device_value)]:
            execute_tcdel(device_option)

            assert (
                SubprocessRunner(
                    " ".join(
                        [
                            Tc.Command.TCSET,
                            device_option,
                            "--delay 100ms --rate 50k --network 192.168.1.2 --change",
                        ]
                    )
                ).run()
                == 0
            )

            assert (
                SubprocessRunner(
                    " ".join(
                        [
                            Tc.Command.TCSET,
                            device_option,
                            "--delay 100ms --rate 50k --network 192.168.1.3 --change",
                        ]
                    )
                ).run()
                == 0
            )

            execute_tcdel(device_option)

    def test_normal(self, device_value):
        if device_value is None:
            pytest.skip("device is null")

        for device_option in [device_value, "--device {}".format(device_value)]:
            assert (
                SubprocessRunner(
                    " ".join(
                        [
                            Tc.Command.TCSET,
                            device_option,
                            "--delay 100ms --rate 50k --network 192.168.1.2 --overwrite",
                        ]
                    )
                ).run()
                == 0
            )

            assert (
                SubprocessRunner(
                    " ".join(
                        [
                            Tc.Command.TCSET,
                            device_option,
                            "--delay 200.0ms",
                            "--delay-distro 20",
                            "--rate 100k",
                            "--loss 0.01%",
                            "--duplicate 5%",
                            "--reorder 2%",
                            "--network 192.168.1.3",
                            "--add",
                        ]
                    )
                ).run()
                == 0
            )

            runner = SubprocessRunner("{:s} {:s}".format(Tc.Command.TCSHOW, device_option))
            runner.run()

            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                            "dst-network=192.168.1.2/32, protocol=ip": {
                                "filter_id": "800::800",
                                "delay": "100.0ms",
                                "rate": "50Kbps"
                            },
                            "dst-network=192.168.1.3/32, protocol=ip": {
                                "filter_id": "800::801",
                                "delay": "200.0ms",
                                "loss": "0.01%",
                                "duplicate": "5%",
                                "delay-distro": "20.0ms",
                                "rate": "100Kbps",
                                "reorder": "2%"
                            }
                        },
                        "incoming": {}
                    }
                }"""
            )

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)
            assert json.loads(runner.stdout) == json.loads(expected)

            assert (
                SubprocessRunner(
                    " ".join(
                        [
                            Tc.Command.TCSET,
                            device_option,
                            "--delay 300ms",
                            "--delay-distro 30",
                            "--rate 200k",
                            "--loss 0.02%",
                            "--duplicate 5.5%",
                            "--reorder 0.2%",
                            "--network 192.168.1.3",
                            "--change",
                        ]
                    )
                ).run()
                == 0
            )

            runner = SubprocessRunner("{:s} {:s}".format(Tc.Command.TCSHOW, device_option))
            runner.run()

            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                            "dst-network=192.168.1.2/32, protocol=ip": {
                                "filter_id": "800::800",
                                "delay": "100.0ms",
                                "rate": "50Kbps"
                            },
                            "dst-network=192.168.1.3/32, protocol=ip": {
                                "filter_id": "800::801",
                                "delay": "300.0ms",
                                "loss": "0.02%",
                                "duplicate": "5.5%",
                                "delay-distro": "30.0ms",
                                "rate": "200Kbps",
                                "reorder": "0.2%"
                            }
                        },
                        "incoming": {}
                    }
                }"""
            )

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)
            assert json.loads(runner.stdout) == json.loads(expected)

            # finalize ---
            execute_tcdel(device_option)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import re


def __null_line_strip(line):
    return line


def __line_strip(line):
    return line.strip()


def split_line_list(line_list, re_block_separator=None, is_include_match_line=False, is_strip=True):
    block_list = []
    block = []
    strip_func = __line_strip if is_strip else __null_line_strip

    if not re_block_separator:
        re_block_separator = re.compile("^$")

    for line in line_list:
        line = strip_func(line)

        if re_block_separator.search(line):
            if block:
                block_list.append(block)

            block = []
            if is_include_match_line:
                block.append(line)
            continue

        block.append(line)

    if block:
        block_list.append(block)

    return block_list



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import argparse
from textwrap import dedent

import logbook

from ._const import TcCommandOutput, TrafficDirection


class ArgparseWrapper(object):
    """
    Wrapper class for argparse
    """

    def __init__(self, version, description=""):
        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=description,
            epilog=dedent(
                """\
                Documentation: https://tcconfig.rtfd.io/
                Issue tracker: https://github.com/thombashi/tcconfig/issues
                """
            ),
        )
        self.parser.add_argument(
            "-V", "--version", action="version", version="%(prog)s {}".format(version)
        )
        self._add_tc_command_arg_group()
        self._add_log_level_argument_group()

        group = self.parser.add_argument_group("Debug")
        group.add_argument(
            "--debug-query", action="store_true", default=False, help="for debug print."
        )
        group.add_argument(
            "--stacktrace",
            dest="is_output_stacktrace",
            action="store_true",
            default=False,
            help="""print stack trace for debug information.
            --debug option required to see the debug print.
            """,
        )

    def add_routing_group(self):
        group = self.parser.add_argument_group("Routing")
        group.add_argument(
            "--direction",
            choices=TrafficDirection.LIST,
            default=TrafficDirection.OUTGOING,
            help="""the direction of network communication that imposes traffic control.
            'incoming' requires ifb kernel module and Linux kernel 2.6.20 or later.
            (default = %(default)s)
            """,
        )
        group.add_argument(
            "--network",
            "--dst-network",
            dest="dst_network",
            help="""specify destination IP-address/network that applies traffic control.
            defaults to any.""",
        )
        group.add_argument(
            "--src-network",
            help="""specify source IP-address/network that applies traffic control.
            defaults to any.
            this option has no effect when executing with "--direction incoming" option.
            note: this option required to execute with the --iptables option when using tbf
            algorithm.
            """,
        )
        group.add_argument(
            "--port",
            "--dst-port",
            dest="dst_port",
            type=int,
            help="specify destination port number that applies traffic control. defaults to any.",
        )
        group.add_argument(
            "--src-port",
            type=int,
            help="specify source port number that applies traffic control. defaults to any.",
        )
        group.add_argument(
            "--ipv6",
            dest="is_ipv6",
            action="store_true",
            default=False,
            help="apply traffic control to IPv6 packets rather than IPv4.",
        )

        return group

    def _add_log_level_argument_group(self):
        dest = "log_level"

        group = self.parser.add_mutually_exclusive_group()
        group.add_argument(
            "--debug",
            dest=dest,
            action="store_const",
            const=logbook.DEBUG,
            default=logbook.INFO,
            help="for debug print.",
        )
        group.add_argument(
            "--quiet",
            dest=dest,
            action="store_const",
            const=logbook.NOTSET,
            default=logbook.INFO,
            help="suppress execution log messages.",
        )

        return group

    def add_docker_group(self, is_add_srcdst=True):
        group = self.parser.add_argument_group("Docker")
        group.add_argument(
            "--docker",
            dest="use_docker",
            action="store_true",
            default=False,
            help="apply traffic control to a docker container.",
        )
        if is_add_srcdst:
            group.add_argument("--src-container", help="specify source container id or name.")
            group.add_argument("--dst-container", help="specify destination container id or name.")

        return group

    def _add_tc_command_arg_group(self):
        group = self.parser.add_mutually_exclusive_group()
        group.add_argument(
            "--tc-command",
            dest="tc_command_output",
            action="store_const",
            const=TcCommandOutput.STDOUT,
            default=TcCommandOutput.NOT_SET,
            help="""
            display tc commands to be executed and exit.
            these commands are not actually executed.
            """,
        )

        group.add_argument(
            "--tc-script",
            dest="tc_command_output",
            action="store_const",
            const=TcCommandOutput.SCRIPT,
            default=TcCommandOutput.NOT_SET,
            help="""
            generate a shell script file that described tc commands.
            this tc script execution result nearly equivalent with the tcconfig command.
            the script can be executed without tcconfig package installation.
            """,
        )



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import pytest
import simplejson as json
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc

from .common import execute_tcdel, print_test_result


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcdel(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--runxfail --device=<test device>"
    """

    def test_normal_ipv4(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        for device_option in [[device_value], ["--device", device_value]]:
            execute_tcdel(device_value)

            tcshow_cmd = [Tc.Command.TCSHOW] + device_option

            assert (
                SubprocessRunner(
                    [Tc.Command.TCSET]
                    + device_option
                    + [
                        "--delay",
                        "10",
                        "--delay-distro",
                        "2",
                        "--loss",
                        "0.01",
                        "--duplicate",
                        "0.5",
                        "--reorder",
                        "0.2",
                        "--rate",
                        "0.25K",
                        "--network",
                        "192.168.0.10",
                        "--port",
                        "8080",
                        "--overwrite",
                    ]
                ).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCSET]
                    + device_option
                    + [
                        "--delay",
                        "1",
                        "--loss",
                        "1",
                        "--rate",
                        "100M",
                        "--network",
                        "192.168.1.0/24",
                        "--add",
                    ]
                ).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCSET]
                    + device_option
                    + [
                        "--delay",
                        "10",
                        "--delay-distro",
                        "2",
                        "--rate",
                        "500K",
                        "--direction",
                        "incoming",
                    ]
                ).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCSET]
                    + device_option
                    + [
                        "--delay",
                        "1",
                        "--loss",
                        "0.02",
                        "--duplicate",
                        "0.5",
                        "--reorder",
                        "0.2",
                        "--rate",
                        "0.1M",
                        "--network",
                        "192.168.11.0/24",
                        "--port",
                        "80",
                        "--direction",
                        "incoming",
                        "--add",
                    ]
                ).run()
                == 0
            )

            runner = SubprocessRunner(tcshow_cmd)
            runner.run()
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                        "dst-network=192.168.0.10/32, dst-port=8080, protocol=ip": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "loss": "0.01%",
                                "duplicate": "0.5%",
                                "reorder": "0.2%",
                                "rate": "248bps",
                                "delay-distro": "2.0ms"
                            },
                            "dst-network=192.168.1.0/24, protocol=ip": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "1%",
                                "rate": "100Mbps"
                            }
                        },
                        "incoming": {
                            "dst-network=192.168.11.0/24, dst-port=80, protocol=ip": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "0.02%",
                                "duplicate": "0.5%",
                                "reorder": "0.2%",
                                "rate": "100Kbps"
                            },
                            "protocol=ip": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "delay-distro": "2.0ms",
                                "rate": "500Kbps"
                            }
                        }
                    }
                }"""
            )

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)

            assert json.loads(runner.stdout) == json.loads(expected)

            tcdel_proc = SubprocessRunner(
                [Tc.Command.TCDEL] + device_option + ["--network", "192.168.1.0/24"]
            )
            assert tcdel_proc.run() == 0, tcdel_proc.stderr
            tcdel_proc = SubprocessRunner(
                [Tc.Command.TCDEL]
                + device_option
                + ["--network", "192.168.11.0/24", "--port", "80", "--direction", "incoming"]
            )
            assert tcdel_proc.run() == 0, tcdel_proc.stderr

            runner = SubprocessRunner(tcshow_cmd)
            runner.run()
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                            "dst-network=192.168.0.10/32, dst-port=8080, protocol=ip": {
                                "delay": "10.0ms",
                                "loss": "0.01%",
                                "duplicate": "0.5%",
                                "filter_id": "800::800",
                                "delay-distro": "2.0ms",
                                "rate": "248bps",
                                "reorder": "0.2%"
                            }
                        },
                        "incoming": {
                            "protocol=ip": {
                                "delay": "10.0ms",
                                "rate": "500Kbps",
                                "delay-distro": "2.0ms",
                                "filter_id": "800::800"
                            }
                        }
                    }
                }"""
            )

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)
            assert json.loads(runner.stdout) == json.loads(expected)

            assert (
                SubprocessRunner([Tc.Command.TCDEL] + device_option + ["--id", "800::800"]).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCDEL]
                    + device_option
                    + ["--id", "800::800", "--direction", "incoming"]
                ).run()
                == 0
            )

            runner = SubprocessRunner(tcshow_cmd)
            runner.run()
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
            "outgoing": {},
            "incoming": {}
        }
    }"""
            )

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)
            assert json.loads(runner.stdout) == json.loads(expected)

            # finalize ---
            execute_tcdel(device_value)

    def test_normal_ipv6(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        for device_option in [[device_value], ["--device", device_value]]:
            execute_tcdel(device_value)

            tcshow_cmd = [Tc.Command.TCSHOW] + device_option

            proc = SubprocessRunner(
                [Tc.Command.TCSET]
                + device_option
                + [
                    "--delay",
                    "10",
                    "--delay-distro",
                    "2",
                    "--loss",
                    "0.01",
                    "--duplicate",
                    "5",
                    "--reorder",
                    "2",
                    "--rate",
                    "0.25K",
                    "--network",
                    "::1",
                    "--port",
                    "8080",
                    "--overwrite",
                    "--ipv6",
                ]
            )
            print(proc.command)
            assert proc.run() == 0
            assert (
                SubprocessRunner(
                    [Tc.Command.TCSET]
                    + device_option
                    + [
                        "--delay",
                        "1",
                        "--loss",
                        "1",
                        "--rate",
                        "100M",
                        "--network",
                        "2001:db00::0/24",
                        "--add",
                        "--ipv6",
                    ]
                ).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCSET]
                    + device_option
                    + [
                        "--delay",
                        "10",
                        "--delay-distro",
                        "2",
                        "--rate",
                        "500K",
                        "--direction",
                        "incoming",
                        "--ipv6",
                    ]
                ).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCSET]
                    + device_option
                    + [
                        "--delay",
                        "1",
                        "--loss",
                        "0.02",
                        "--duplicate",
                        "5",
                        "--reorder",
                        "2",
                        "--rate",
                        "0.1M",
                        "--network",
                        "2001:db00::0/25",
                        "--port",
                        "80",
                        "--direction",
                        "incoming",
                        "--add",
                        "--ipv6",
                    ]
                ).run()
                == 0
            )

            runner = SubprocessRunner(tcshow_cmd + ["--ipv6"])
            runner.run()

            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                            "dst-network=::1/128, dst-port=8080, protocol=ipv6": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "loss": "0.01%",
                                "duplicate": "5%",
                                "reorder": "2%",
                                "rate": "248bps",
                                "delay-distro": "2.0ms"
                            },
                            "dst-network=2001:db00::/24, protocol=ipv6": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "1%",
                                "rate": "100Mbps"
                            }
                        },
                        "incoming": {
                            "dst-network=2001:db00::/25, dst-port=80, protocol=ipv6": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "0.02%",
                                "duplicate": "5%",
                                "reorder": "2%",
                                "rate": "100Kbps"
                            },
                            "protocol=ipv6": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "rate": "500Kbps",
                                "delay-distro": "2.0ms"
                            }
                        }
                    }
                }"""
            )

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)
            assert json.loads(runner.stdout) == json.loads(expected)

            assert (
                SubprocessRunner(
                    [Tc.Command.TCDEL] + device_option + ["--network", "2001:db00::0/24", "--ipv6"]
                ).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCDEL]
                    + device_option
                    + [
                        "--network",
                        "2001:db00::0/25",
                        "--port",
                        "80",
                        "--direction",
                        "incoming",
                        "--ipv6",
                    ]
                ).run()
                == 0
            )

            runner = SubprocessRunner(tcshow_cmd)
            runner.run()
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                            "protocol=ipv6": {
                                "delay": "10.0ms",
                                "loss": "0.01%",
                                "duplicate": "5%",
                                "filter_id": "800::800",
                                "delay-distro": "2.0ms",
                                "rate": "248bps",
                                "reorder": "2%"
                            }
                        },
                        "incoming": {
                            "protocol=ipv6": {
                                "delay": "10.0ms",
                                "rate": "500Kbps",
                                "delay-distro": "2.0ms",
                                "filter_id": "800::800"
                            }
                        }
                    }
                }"""
            )

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)
            assert json.loads(runner.stdout) == json.loads(expected)

            assert (
                SubprocessRunner(
                    [Tc.Command.TCDEL] + device_option + ["--id", "800::800", "--ipv6"]
                ).run()
                == 0
            )
            assert (
                SubprocessRunner(
                    [Tc.Command.TCDEL]
                    + device_option
                    + ["--id", "800::800", "--direction", "incoming", "--ipv6"]
                ).run()
                == 0
            )

            runner = SubprocessRunner(tcshow_cmd)
            runner.run()
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {},
                        "incoming": {}
                    }
                }"""
            )

            # finalize ---
            execute_tcdel(device_value)

    def test_abnormal(self):
        assert execute_tcdel("not_exist_device") != 0



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import pytest
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc

from .common import execute_tcdel


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


class Test_tcset_iface_speed(object):
    @pytest.mark.parametrize(["speed"], [[1], [-1]])  # para-virtualized network driver
    def test_smoke_speed(self, monkeypatch, device_option, speed):
        if device_option is None:
            pytest.skip("device option is null")

        for tc_target in [device_option, "--device {}".format(device_option)]:
            execute_tcdel(tc_target)
            monkeypatch.setattr("tcconfig._network._read_iface_speed", lambda x: speed)

            command_list = [Tc.Command.TCSET, tc_target, "--rate {:s}".format("1kbps")]
            assert SubprocessRunner(" ".join(command_list)).run() == 0

            # finalize ---
            execute_tcdel(tc_target)

    @pytest.mark.parametrize(["rate"], [["0kbps"], ["99Gbps"]])
    def test_abnormal(self, monkeypatch, device_option, rate):
        if device_option is None:
            pytest.skip("device option is null")

        for tc_target in [device_option, "--device {}".format(device_option)]:
            execute_tcdel(tc_target)
            monkeypatch.setattr("tcconfig._network._read_iface_speed", lambda x: "1")

            command_list = [Tc.Command.TCSET, tc_target, "--rate {:s}".format(rate)]
            assert SubprocessRunner(" ".join(command_list)).run() != 0

            # finalize ---
            execute_tcdel(tc_target)
